/* hexprog.cc   Benutzung des Bootloaders fuer Atmega2560   letzte nderung: 7.10.2012 */
#define VERSION "Version 0.2"
/*
History:
29.9.2012	Erstellung aus ser.cc (RP)

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <termios.h>
#include <unistd.h>

/*************************** kleinkram ***************************/
int getline(FILE *fp,char *s,int lim)
{		/* liest eine Textzeile oder maximal lim Zeichen */
		/* und ersetzt den Zeilentrenner durch 0         */
 int c;
 while(--lim && (c=getc(fp))!=EOF && c!='\n')
    if(c!='\r') *s++ = c;
 *s='\0';
 return (c!=EOF);	/* TRUE wenn erfolgreich, FALSE wenn Fileende */
}

int ascitodigit(int c)
{		/* wandelt ein ASCI-Zeichen in eine 1-stellige Hex-Zahl */
 if(!isxdigit(c)) return 0;
 if(islower(c)) c=toupper(c)-'0';
 else c-='0';
 if(c>9) c += '0'-'A'+10;
 return c;
}

/**************** Routinen zur Parameterauswertung ********************/
#define MAXARG 3
static char argflag[128];
void setargflags(char *s)
{
 int c;
 while((c= *s++)!=0)
  {if(c>='a' && c<='z')  c -= 'a'-'A';
   argflag[c&127]=1;
  }
}

/*********************** Serielle Schnittstelle ***********************/
//static char sername[80]="/dev/ttyS0";
static char sername[80]="/dev/ttyUSB1";
//static char sername[80]="/dev/tty.serial-0001"; //fuer aelteren Mac
//static char sername[80]="/dev/tty.usb2uart"; //fuer neuen Mac
static int baudrate=500000;
//static int baudrate=9600; //test
static int fdr,fdw;

static int baudcodliste[]={
   //50,B50, 75,B75, 110,B110, 134,B134, 150,B150,
   300,B300, 600,B600, 1200,B1200, 1800,B1800, 2400,B2400, 4800,B4800,
   9600,B9600, 19200,B19200, 38400,B38400,
   57600,B57600,
   115200,B115200,
   230400,B230400,
   460800,B460800, 500000,B500000,
/*
   576000,B576000,
   921600,B921600, 1000000,B1000000, 1152000,B1152000, 1500000,B1500000,
   2000000,B2000000, 2500000,B2500000, 3000000,B3000000, 3500000,B3500000,
   4000000,B4000000,
*/
   0,0};

int ibaud(int x) //Baudcode in Baudrate umrechnen
{
 int *co,a;
 for(co=baudcodliste;(a= *co++)!=0;co++)
    if(x== *co) return a;
 printf("Fehler: unbekannter Baudratecode %d\n",x);
 return 0;
}

int baudi(int x) //Baudrate in Baudcode umrechnen
{
 int *co,b,min=0;
 for(co=baudcodliste;(b= *co++)>0;co++)
     if(x==b) return *co;
     else if(x>b) {min= *co;}
 printf("Warnung: unbekannte Baudrate %d, auf %d abgerundet\n",x,min);
 return min;
}

void noecho(int fd)
{
 struct termios buf;
 if(tcgetattr(fd,&buf)<0)
   {printf("tcgetattr() Error\n"); return;}
 buf.c_lflag &= ~ECHO;
 if(baudrate!=0)
     {cfsetispeed(&buf,baudi(baudrate));
      cfsetospeed(&buf,baudi(baudrate));
     }
 if(tcsetattr(fd,TCSAFLUSH,&buf)<0)
   {printf("tcsetattr() Error\n"); return;}
 int a=cfgetispeed(&buf),b=cfgetospeed(&buf);
 if(argflag['V']) fprintf(stderr,"Baudraten: %d %d\n",ibaud(a),ibaud(b));
}

void seriell_open()
{
 fdr=open(sername,O_RDWR);
 if(fdr<0) {fprintf(stderr,"Fehler: fdr = %d\n",fdr); exit(0);}
	   //wenn dieser Fehler kommt sollte /dev/ttyS0 ueberprueft werden
	   //Zugriffsberechtigung freigeben: chmod a+rw /dev/ttyS0
	   //oder vielleicht /dev/ttyS1
 noecho(fdr);
 fdw=fdr;
}

void seriell_close()
{
 close(fdr);
}

void seriell_readline(char *str,int max)
{
 int i;
 char *s,c;
 for(i=1,s=str;i<max;)
   {read(fdr,&c,1);
    if(c=='\n') break;
    if(c==0) fprintf(stderr,"\\0");//test
    else
      {*s++ = c; i++;}
   }
 if(s!=str && s[-1] == '\r')
   {--s;
    fprintf(stderr,"CR empfangen\n");//test
   }
 *s=0;
}

int seriell_readbyte()
{
 char c;
 read(fdr,&c,1);
 return c;
}

void seriell_writeline(const char *str)
{
 int n=strlen(str);
 char *wbuf=new char[n+3];
 //if(argflag['C']) {sprintf(wbuf,"%s\r\n",str); n++;}
 //else
 sprintf(wbuf,"%s\n",str);
 write(fdw,wbuf,n+1);
 delete wbuf;
}

/************************* Hauptprogramm ******************************/
int checksumtest(const char *s);

int main(int argc,char *argv[])
{
 char name[80],name2[80],name3[80],zeile[80],rzeile[80];
 int i,j,c=0;
 FILE *fp=NULL;
 name[0]=name2[0]=name3[0]=0;
 for(j=0,i=1;i<argc;i++)
	{if((c= *argv[i])=='-' || c=='?') setargflags(argv[i]);
	 else	{if(++j==1) strcpy(name2,argv[i]);
	         else if(j==2) strcpy(name3,argv[i]);
	         else if(j==3) strcpy(name,argv[i]);
	}	}
 if(argflag['?'] || j>MAXARG)
	{printf("hexprog  %s\n",VERSION);
	 printf("Anwendung: hexprog [/dev/ttyUSB1] [Baudrate] -<Flags> [name.hex]\n");
	 printf("    voreingestellt: %s   %d\n",sername,baudrate);
	 printf(" <Flags>: c=Verbindung pruefen\n");
	 printf("          wf=write flash, Brennen des Programms\n");
	 printf("          we=write EEPROM, Brennen des EEPROMs\n");
	 printf("          rf=read flash, einlesen und als flash.raw speichern\n");
	 printf("          re=read EEPROM, einlesen und als eeprom. speichern\n");
	 printf("          v=verbose, mit Testausgaben\n");
	 //printf("          x=senden ohne Startzeichen abzuwarten\n");//test
	 printf("Beim Controller muss nach dem Start von hexprog ein Reset ausgeloest werden.\n");
	 printf("Beispiel: EEPROM einlesen und in eeprom.hex speichern:\n");
	 printf("  hexprog /dev/ttyUSB1 -vre >eeprom.hex\n");
	 exit(0);
	}
 if(j>=1)
     {if(isdigit(*name2)) {sscanf(name2,"%d",&baudrate); *name2=0;}
      else if(isdigit(*name3)) {sscanf(name3,"%d",&baudrate); *name3=0;}
      else if(isdigit(*name)) {sscanf(name,"%d",&baudrate); *name=0;}
      if(argflag['V']) fprintf(stderr,"Baudrate = %d\n",baudrate);
      if(name2[0]=='/' || strncmp(name2,"COM",3)==0) {strcpy(sername,name2); *name2=0;}
      else if(name3[0]=='/' || strncmp(name3,"COM",3)==0) {strcpy(sername,name3); *name3=0;}
      else if(name[0]=='/' || strncmp(name,"COM",3)==0) {strcpy(sername,name); *name=0;}
      if(argflag['V']) fprintf(stderr,"SerielleSchnittstelle = \"%s\"\n",sername);
      if(name[0]==0)
       {if(name2[0]!=0) strcpy(name,name2);
	else if(name3[0]!=0) strcpy(name,name3);
       }
      if(argflag['V']) fprintf(stderr,"Name = \"%s\"\n",name);
     }
 if(argflag['V']) fprintf(stderr,"Serielle Schnittstelle '%s' wird geoeffnet\n",sername);
 seriell_open();

 fprintf(stderr,"Jetzt Reset vom Atmega2560 ausloesen. Reset-Knopf oder Spannung Aus/Ein-schalten.\n");
 sprintf(zeile,"%s","StartBootloader"); //Antwort auf Startzeichen vorbereiten

 //if(argflag['X']) {c='X'; sleep(1);} else //test: statt Startbyte empfangen 1 Sekunde warten
 //c=seriell_readbyte(); //Startzeichen empfangen, funktioniert so nicht!

 {seriell_readline(rzeile,80); c=rzeile[0];}  //Startzeichen empfangen
 seriell_writeline(zeile); //und sofort Antwort senden.
 if(argflag['V'])
  {fprintf(stderr,"empfangenes Startzeichen: '%c'\n",c);
   fprintf(stderr,"gesendet: \"%s\"\n",zeile);
  }
 seriell_readline(rzeile,80); if(argflag['V']) fprintf(stderr,"empfangen: \"%s\"\n",rzeile);
 if(argflag['C']) //nur Verbindungstest machen
  {
   printf("Verbindungstest:\n");
   sprintf(zeile,"%s","check");
   seriell_writeline(zeile);    printf("gesendet:  \"%s\"\n",zeile);
   seriell_readline(rzeile,80); printf("empfangen: \"%s\"\n",rzeile);
   sprintf(zeile,"%s","Verbindungstest von hexprog.cc zum Atmega2560");
   seriell_writeline(zeile);    printf("gesendet:  \"%s\"\n",zeile);
   seriell_readline(rzeile,80); printf("empfangen: \"%s\"\n",rzeile);
   sprintf(zeile,"%s","reset");
   seriell_writeline(zeile);    printf("gesendet:  \"%s\"\n",zeile);
   //seriell_readline(rzeile,80); printf("empfangen: \"%s\"\n",rzeile);
  }
 else if(argflag['W'] && argflag['F'] && name[0]!=0)
  {
   //if(argflag['V'])
     printf("Datei \"%s\" wuerde ins Flash gebrannt\n",name);
  }
 else if(argflag['R'] && argflag['F'])
  {
   //if(argflag['V'])
     printf("Flash wuerde eingelesen\n");
  }
 else if(argflag['W'] && argflag['E'] && name[0]!=0)
  {
   int nfehler=1;
   //if(argflag['V'])
   fprintf(stderr,"Datei \"%s\" wird ins EEPROM gebrannt\n",name);
   fp=fopen(name,"r");
   if(fp==NULL)
    {fprintf(stderr,"Error: cant open \"%s\"\n",name); nfehler=0;}
   else if(strcmp(rzeile,"ok")==0)
    {
     sprintf(zeile,"%s","writeeprom");
     if(argflag['V']) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     seriell_readline(rzeile,80);
     if(strcmp(rzeile,"ok")==0)
      {nfehler=0;
       while(getline(fp,zeile,80))
	{
	 do
	  {if(argflag['V']) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
	   seriell_writeline(zeile);
	   seriell_readline(rzeile,80);
	  }
	 while(strcmp(rzeile,"err-checksum")==0 && ++nfehler<5); //Bei Pruefsummenfehler nochmals senden
	 if(strcmp(rzeile,"ok")!=0) break;
	 nfehler=0;
	}
       if(nfehler==0)
	{
	 seriell_writeline("ok");
	 if(argflag['V']) fprintf(stderr,"gesendet:  \"ok\"\n");
	}
      }
    }
   if(nfehler>0) fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else if(argflag['R'] && argflag['E'])
  {
   //if(argflag['V'])
    fprintf(stderr,"EEPROM wird eingelesen\n");
   if(strcmp(rzeile,"ok")==0)
    {
     sprintf(zeile,"%s","readeprom");
     if(argflag['V']) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     for(;;)
      {int checksum=0;
       seriell_readline(rzeile,80);
       if(rzeile[0]==':' && (checksum=checksumtest(&rzeile[1]))==0)
	{
	 printf("%s\n",rzeile); //Die Hex-Daten auf Standardausgabe oder mit ">name.hex" in die Datei
	 seriell_writeline("ok"); //erfolgreicher Empfang mit "ok" bestaetigen
	}
       else if(strcmp(rzeile,"ok")==0) //Ende der Daten wenn "ok" empfangen wird
	{
	 fprintf(stderr,"EEPROM erfolgreich eingelesen.\n");
	 break; //erfolgreich fertig
	}
       else if(checksum!=0)
	{
	 fprintf(stderr,"checksum-error - try again\n");
	 seriell_writeline("err-checksum");
	}
       else
	{
	 fprintf(stderr,"Fehler: \"%s\"\n",rzeile); break; //Abbrechen bei unbekanntem Fehler
	}
      }
    }
   else fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else
  {
   fprintf(stderr,"Fehlendes oder falsches Kommando. Nichts gemacht.\n");
  }

 /*
   if(*zeile!=0) seriell_writeline(zeile);
   seriell_readline(rzeile,80);
 */

 if(argflag['V']) fprintf(stderr,"Serielle Schnittstelle wird geschlossen\n");
 seriell_close();
 return 0;
}/* ende von main */

int checksumtest(const char *s)
{
 int pruefsumme=0,c1,c2;
 while((c1= *s++)!=0 && (c2= *s++)!=0)
  {
   pruefsumme += (ascitodigit(c1)<<4)+ascitodigit(c2);
  }
 return (pruefsumme&0xFF);
}
