/* uart.c       letzte Aenderung: 29.9.2012
UART Ansteuerung
teilweise von mikrocontroller.net kopiert:
http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial/Der_UART

 Version: 0.02
 Prozessor: ATmega32, ATmega8, Atmega2560
 Schaltung:
   Senden:    PD1 (TxD) oder TxD3 beim Atmega2560
   Empfangen: PD0 (RxD) oder RxD3 beim Atmega2560

 Autor: Rolf Pfister
 Copyright: Freeware
 History:
 9.8.2010	Erstellung
 11.11.2010	erweitert auf vollstaendiges Programm
 29.9.2012      Anpassung fuer Atmega2560, Namen geaendert von uart.cc auf uart.c

Verwendung im Hauptprogramm:
#include <inttypes.h>
#include <avr/io.h>
#define ATMEGA2560  //fuer andere Controller auskommentieren
#include "uart.c"

*/
//#define HAUPTPROGRAMM  //auskommentieren wenn wie oben eingefuegt

#ifdef HAUPTPROGRAMM
#define F_CPU 16000000UL
#include <inttypes.h>
#include <avr/io.h>
#endif
#ifndef bool
#define bool int8_t
#endif

#ifndef F_CPU
#warning "F_CPU war noch nicht definiert, wird nun nachgeholt mit 16000000"
#define F_CPU 16000000UL  // Systemtakt in Hz - Definition als unsigned long
                          // Ohne ergeben sich unten Fehler in der Berechnung
#endif

#ifndef BAUD
//#define BAUD 9600UL      // Baudrate
//#define BAUD 38400UL      // Baudrate
#define BAUD 500000UL      // Baudrate
#endif

// Berechnungen
#define UBRR_VAL ((F_CPU+BAUD*8)/(BAUD*16)-1)   // clever runden
#define BAUD_REAL (F_CPU/(16*(UBRR_VAL+1)))     // Reale Baudrate
#define BAUD_ERROR ((BAUD_REAL*1000)/BAUD) // Fehler in Promille, 1000 = kein Fehler.
 
#if ((BAUD_ERROR<990) || (BAUD_ERROR>1010))
  #error Systematischer Fehler der Baudrate groesser 1% und damit zu hoch! 
#endif 

void uart_init()
{
#ifdef ATMEGA2560

 UBRR3 = UBRR_VAL;
 UCSR3B = (1<<RXEN3) | (1<<TXEN3); // UART RX und TX einschalten
 //UCSR3C = (1<<USBS3)|(3<<UCSZ30); // Asynchron 8N2 Atmega2560
 UCSR3C = (3<<UCSZ30); // Asynchron 8N1 Atmega2560

#else
 //UBRRH = UBRR_VAL >> 8;
 //UBRRL = UBRR_VAL & 0xFF;
 UBRR = UBRR_VAL; //sollte mit neueren Compilern auch gehen

 //UCSRB |= (1<<TXEN); // UART TX einschalten
 //UCSRB |= (1<<RXEN); // UART RX einschalten
 UCSRB = (1<<RXEN) | (1<<TXEN); // UART RX und TX einschalten

 //UCSRC |= (1<<URSEL)|(1<<UCSZ1)|(1<<UCSZ0); // Asynchron 8N1 Atmega16
 //UCSRC = (1<<URSEL)|(1<<USBS)|(3<<UCSZ0); // 8data, 2stop bit (8N2) Atmega32
 //UCSRC = (1<<URSEL)|(2<<UPM0)|(3<<UCSZ0); // 8data, 1stop, Even Parity
 //UCSRC = (1<<URSEL)|(3<<UPM0)|(3<<UCSZ0); // 8data, 1stop, Odd Parity
 UCSRC = (1<<URSEL)|(3<<UCSZ0); // Asynchron 8N1 Atmega32 oder Atmega8
#endif
}

void uart_putc(char c)
{
#ifdef ATMEGA2560
 while (!(UCSR3A & (1<<UDRE3))) ;/* warten bis Senden moeglich */
 UDR3 = c;                      /* sende Zeichen */
#else
 while (!(UCSRA & (1<<UDRE))) ;/* warten bis Senden moeglich */
 UDR = c;                      /* sende Zeichen */
#endif
}

bool uart_getcheck()
{
#ifdef ATMEGA2560
 return (UCSR3A & (1<<RXC3)); //true wenn ein Zeichen verfuegbar
#else
 return (UCSRA & (1<<RXC)); //true wenn ein Zeichen verfuegbar
#endif
}

char uart_getc(void)
{
#ifdef ATMEGA2560
 while (!(UCSR3A & (1<<RXC3))) ;// warten bis Zeichen verfuegbar
 return UDR3; // Zeichen aus UDR an Aufrufer zurueckgeben
#else
 while (!(UCSRA & (1<<RXC))) ;// warten bis Zeichen verfuegbar
 return UDR; // Zeichen aus UDR an Aufrufer zurueckgeben
#endif
}

void uart_write(const char *s)
{
 char c;
 while((c= *s++)!=0) uart_putc(c);
}

#ifdef HAUPTPROGRAMM
#include <stdio.h>

int main()
{
 char c,text[120],zeile[80];
 int8_t i=0;
 uart_init();
 while(1)
  {
   //uart_write("Hallo Welt\n");
   if(uart_getcheck())
     {
      c=uart_getc();
      //uart_putc(c);
      //uart_write(" uart_getc() erfolgreich\n");
      if(c=='\n')
      //if(c=='x')
       {
	zeile[i]=0;
	sprintf(text,"Zeile erfolgreich empfangen: '%s'\n",zeile);
	uart_write(text);
	zeile[i=0]=0;
       }
      else if(i<80-1 && c>=' ') zeile[i++]=c;
     }
  }
 return 0;
}
#endif
