/* uart3.h       letzte Aenderung: 27.10.2012
UART Ansteuerung
teilweise von mikrocontroller.net kopiert:
http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial/Der_UART

 Version: 0.02
 Prozessor: Atmega2560
 Schaltung:
   Senden:    TxD3 (PJ1)
   Empfangen: RxD3 (PJ0)

 Autor: Rolf Pfister
 Copyright: Freeware
 History:
 10.10.2012	Erstellung aus uart.c und Anpassung als Include-datei

Verwendung im Hauptprogramm:
#include "uart3.h"

*/
#ifndef UART3_H
#define UART3_H

#ifndef F_CPU
#warning "F_CPU war noch nicht definiert, wird nun nachgeholt mit 16000000"
#define F_CPU 16000000UL  // Systemtakt in Hz - Definition als unsigned long
                          // Ohne ergeben sich unten Fehler in der Berechnung
#endif

#ifndef BAUD
//#define BAUD 9600UL      // Baudrate
#define BAUD 500000UL      // Baudrate
#endif

// Berechnungen
#ifdef DOUBLESPEED
#define UBRR_VAL ((F_CPU+BAUD*4)/(BAUD*8)-1)   // clever runden
#define BAUD_REAL (F_CPU/(8*(UBRR_VAL+1)))     // Reale Baudrate
#else
#define UBRR_VAL ((F_CPU+BAUD*8)/(BAUD*16)-1)   // clever runden
#define BAUD_REAL (F_CPU/(16*(UBRR_VAL+1)))     // Reale Baudrate
#endif
#define BAUD_ERROR ((BAUD_REAL*1000)/BAUD) // Fehler in Promille, 1000 = kein Fehler.

#if ((BAUD_ERROR<990) || (BAUD_ERROR>1010))
  #warning Systematischer Fehler der Baudrate groesser 1%
#endif 

#if ((BAUD_ERROR<980) || (BAUD_ERROR>1020))
  #warning Systematischer Fehler der Baudrate groesser 2%
#endif 

#if ((BAUD_ERROR<970) || (BAUD_ERROR>1030))
  #warning Systematischer Fehler der Baudrate groesser 3%
#endif 

#if ((BAUD_ERROR<960) || (BAUD_ERROR>1040))
  #warning Systematischer Fehler der Baudrate groesser 4%
#endif 

#if ((BAUD_ERROR<950) || (BAUD_ERROR>1050))
  #error Systematischer Fehler der Baudrate groesser 5% und damit zu hoch! 
#endif 

// Initialisierung fuer Asynchron 8N1:
#ifdef DOUBLESPEED
#define uart_init() {UBRR3=UBRR_VAL; UCSR3A=(1<<U2X3); UCSR3B=(1<<RXEN3)|(1<<TXEN3); UCSR3C=(3<<UCSZ30);}
#else
#define uart_init() {UBRR3=UBRR_VAL; UCSR3B=(1<<RXEN3)|(1<<TXEN3); UCSR3C=(3<<UCSZ30);}
#endif

// Ein Zeichen senden:
#define uart_putc(c) {while(!(UCSR3A & (1<<UDRE3))) ; UDR3=(c);} //Vorsicht Stolperfalle mit ;

// Pruefen ob ein Zeichen empfangen wurde:
#define uart_getcheck() ((UCSR3A & (1<<RXC3))!=0)

// Empfangenes Zeichen lesen:
#define uart_getc_nowait() (UDR3)

#define uart_puts(s) {const char *t=(s); while((*t)!=0) uart_putc(*t++);} //Vorsicht Stolperfalle mit ;

//Auf Defaultwerte nach Datenblatt zuruecksetzen:
#ifdef DOUBLESPEED
#define uart_off() {UCSR3A=0; UCSR3B=0; UCSR3C = (3<<UCSZ30); UBRR3=0;}
#else
#define uart_off() {UCSR3B = 0; UCSR3C = (3<<UCSZ30); UBRR3 = 0;}
#endif

#endif
