/* uart3soft.c       letzte Aenderung: 26.10.2012
UART Ansteuerung mit Software fuer flexibere Baudraten

 Version: 0.0
 Prozessor: Atmega2560
 Schaltung:
   Senden:    TxD3 (PJ1)
   Empfangen: RxD3 (PJ0)

 Autor: Rolf Pfister
 Copyright: Freeware
 History:
 26.10.2012	Erstellung aus uart.c und Anpassung als Include-datei

Verwendung im Hauptprogramm:
#include "uart3soft.c"

*/
#ifndef F_CPU
#warning "F_CPU war noch nicht definiert, wird nun nachgeholt mit 16000000"
#define F_CPU 16000000UL  // Systemtakt in Hz - Definition als unsigned long
                          // Ohne ergeben sich unten Fehler in der Berechnung
#endif

#ifndef BAUD
//#define BAUD 9600UL      // Baudrate
#define BAUD 500000UL      // Baudrate
#endif

// Berechnungen
#define TAKTEPROBIT ((F_CPU+BAUD/2)/BAUD)   // clever runden
#define BAUD_REAL (F_CPU/TAKTEPROBIT)     // Reale Baudrate
#define BAUD_ERROR ((BAUD_REAL*1000)/BAUD) // Fehler in Promille, 1000 = kein Fehler.
 
#if ((BAUD_ERROR<990) || (BAUD_ERROR>1010))
  #error Systematischer Fehler der Baudrate groesser 1% und damit zu hoch! 
#endif 

// Initialisierung fuer Asynchron 8N1:
/*
void uart_init()
{
 //DDRJ |= (1<<1);  //PJ1 als Ausgang (TxD3)
 //PORTJ &= ~(1<<1);
 //DDRJ &= ~(1<<0); //PJ0 als Eingang (RxD3)
}
*/
//Sparvariante:
#define uart_init() DDRJ = (1<<1)

// Ein Zeichen senden:
#define TAKTEIMLOOP1 17 //noch anpassen: vom .lss auszaehlen
void uart_putc(uint8_t c)
{
 uint8_t i,jbit=1;
 for(i=0;i<10;i++) //mit Start- und Stop-Bit total 10 Bits senden
  {
   PORTJ = (PORTJ & ~(1<<1)) | (jbit<<1);
   jbit = c>>7;
   c <<= 1;
   _delay_us((TAKTEPROBIT-TAKTEIMLOOP1)/16.0);
  }
}

// Pruefen ob ein Zeichen empfangen wurde:
static uint8_t uart_empfang=0;

#define TAKTEIMLOOP2 9  //noch anpassen: vom .lss auszaehlen: Takte in innerer Schlaufe
#define TAKTEIMLOOP3 10 //noch anpassen: vom .lss auszaehlen: Takte in aeusserer Schlaufe
uint8_t uart_getcheck()
{
 int8_t i,j;
 for(i=10;i>0;i--)
  if((PINJ & (1<<0))!=0) //Startbit gesetzt?
   {
    for(j=0;j<9;j++) //Startbit+8 Datenbits einlesen
     {
      uart_empfang<<=1;
      uart_empfang |= (PINJ & 1); //wenn es nicht J0 waere: ((PINJ & (1<<x))>>x);
      _delay_us((TAKTEPROBIT-TAKTEIMLOOP2)/16.0); //warten auf naechstes Bit
     }
    return 1;
   }
  else if(i!=1)
   {
    _delay_us((TAKTEPROBIT/3-TAKTEIMLOOP3)/16.0); //etwas weniger als halbe Bit-Laenge warten
   }
 return 0;
}

// Empfangenes Zeichen lesen:
#define uart_getc_nowait() uart_empfang

#define uart_puts(s) {const char *t=(s); while((*t)!=0) uart_putc(*t++);} //Vorsicht Stolperfalle mit ;

//Auf Defaultwerte zuruecksetzen:
/*
void uart_off()
{
 DDRJ=0;  //alles wieder Eingaenge
 PORTJ=0; //und keine Pullups
}
*/
//Sparvariante:
#define uart_off() DDRJ=0
